#define BUILD_NUMBER \
	"build-6082533"
#define BUILD_NUMBER_NUMERIC \
	6082533
#define BUILD_NUMBER_NUMERIC_STRING \
	"6082533"
#define PRODUCT_BUILD_NUMBER \
	"product-build-63510"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	63510
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"63510"
